#! /usr/local/bin/perl -w

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.



use Getopt::Std;

######################################################################################################
#   GLOBAL VARIABLES AND COMMAND LINE OPTIONS
######################################################################################################

$word_choices_cuis = " /share/nlp/projects/LexDis/data/NLM-WSD.target_word.choices/choices/2007/"; 


######################################################################################################
#   PROCESS COMMAND LINE OPTIONS
######################################################################################################

# Read the command line options
getopts('s:d:i:h', \%opts); 

if(exists $opts{h}) {&usage();}

# Dictionary file to dump data to
if(exists $opts{d}) {
  $dict_file = $opts{d};
} else {
  $dict_file = "Dictionary.txt";
} # if/else


# File containing list of stop phrases
if(exists $opts{s}) {
  $stop_phrase_file = $opts{s};
  $stopwords_defined = 1;
} else {
  $stopwords_defined = 0;
} # if/else


# Version of NLM corpus that has been processed by MetaMap
if(exists $opts{i}) {
  $proc_corpus = $opts{i}."/";
} else {
  $proc_corpus = "./OUT/";
} # if/else


# Hash containing mapping from each term to possible meanings
%dictionary = ();


######################################################################################################
#   READ IN LIST OF STOP PHRASES
######################################################################################################

%stop_phrases = ();
if($stopwords_defined == 1) {
  open(STOPS, $stop_phrase_file) or die "Can't open $stop_phrase_file for reading\n";
  while(<STOPS>) {
    if(/(.+)/) {
      $stop_phrases{$1}++;
    } # if
  } # while
  close(STOPS);
} # if

######################################################################################################
#   START OF SCRIPT PROPER
######################################################################################################

# Read in list of possible CUIs
%gold_standard_cuis = ();

open(PIPE, "ls $word_choice_cuis/*.choices |") or die "Can't open pipe\n";

while(<PIPE>) {
  if(/^(.+\/([a-z\_]+)\.choices)$/) {
    $term = $2;
    $filename = $1;
    open(FILE, $filename) or die "Can't open $filename for reading\n";
    while(<FILE>) {
      if(/^(M[0-9]+)\|[^\|]+\|[^\|]+\|(C[0-9]+)/) {
	$gold_standard_cuis{$term}{$1} = $2;
      } # if
    } # while
  } # if
} # while


# Read through corpus of documents that have been pre-processed with MetaMap
$command = "ls $proc_corpus/* |";
open(PIPE, $command) or die "Can't execute command as pipe: $command\n";
while(<PIPE>) {
  $file = $_;

  %mappings = ();
  %targets = ();
  %context_ids = ();

  $abstract_count = 0;
  $sent_count = 0;
  $target_sent = 0;
  $prev_sent_target = 0;
  $phrase_count = 0;
  $map_count = 0;
  $term_count = 0;

  open(FILE, $file) or die "Can't open $file for reading\n";
  while(<FILE>) {
    if(/^([a-z\_]+)\:([0-9]+\|[0-9]+\.[tiab]+\.[0-9]+\|(M[0-9]|None))/) {
      $abstract_count++;
      $context_ids{$abstract_count} = $1.":".$2;
      $sent_count = 0;

    } elsif(/^Processing /) {

      # Note that this is the start of a new sentence, unless it is phrase containing the target 
      # sentence or the one following it. 
      if($target_sent == 1) {
	$target_sent = 0;
	$prev_sent_target = 1;
      } elsif($prev_sent_target == 1) {
	$prev_sent_target = 0;
      } else {
	$sent_count++;
      } # if/else

    } elsif(/Phrase\: \"([^\"]+)\"/) {
      # Start of new phrase
      $phrase = $1;
      if(exists $stop_phrases{$phrase}) {
	$skip_phrase++;
      } else {
	$skip_phrase = 0;
	
	$phrase_count++;
	$map_count = 0;
      } # if/esle

    } elsif(/\*\* Ambiguous term \*\*/) {
      # Check whether next phrase in the target, if it is then need to store some information
      # and adjust some counts
      $next_phrase = $phrase_count + 1;
      $targets{$abstract_count}{$sent_count}{$next_phrase}++;
      $target_sent = 1;

    } elsif(/Meta Mapping /) {
      unless($skip_phrase == 1) {
	# Start of new mapping
	$map_count++;
	$term_count = 0;
      } # unless

    } elsif(/(C[0-9]+)\:([^\(\[]+)/) {
      unless($skip_phrase == 1) {
	$cui = $1;
	$term = $2;
	$norm_term = &norm_term($term);
	$mappings{$abstract_count}{$sent_count}{$phrase_count}{$map_count}{$term_count}{"term"} = $norm_term;
	$mappings{$abstract_count}{$sent_count}{$phrase_count}{$map_count}{$term_count}{"cui"} = $cui;
	$term_count++;
      } # unless
    } # if/else

  } # while
  close(FILE);


  # Print out text 
  foreach $abstract_count (sort {$a <=> $b} keys %mappings) {
    $context_id = $context_ids{$abstract_count};
    if($context_id =~ /^([a-z\_]+)\:.+\|(M[0-9]+)$/) {
      $ambig_term = $1;
      $sense = $2;
      $correct_cui = $gold_standard_cuis{$ambig_term}{$sense};
    } else { $correct_cui = "None"; } 
    print "Context:$context_ids{$abstract_count}:$correct_cui\n";
    $word_id = 1;

    foreach $sent_count (sort {$a <=> $b} keys %{$mappings{$abstract_count}}) {
      foreach $phrase_count (sort {$a <=> $b} keys %{$mappings{$abstract_count}{$sent_count}}) {

	# Print out text
	foreach $map_count (sort {$a <=> $b} keys %{$mappings{$abstract_count}{$sent_count}{$phrase_count}}) {
	  foreach $term_count (sort {$a cmp $b} keys %{$mappings{$abstract_count}{$sent_count}{$phrase_count}{$map_count}}) {
	    if($map_count == 1) {
	      print "$mappings{$abstract_count}{$sent_count}{$phrase_count}{$map_count}{$term_count}{term}#n#w$word_id#";
	      if(exists $targets{$abstract_count}{$sent_count}{$phrase_count} && $targets{$abstract_count}{$sent_count}{$phrase_count} > 0) {
		print "1 ";
	      } else {
		print "0 ";
	      } # if/else
	      $word_id++;
	    } # if
	    $dictionary{$mappings{$abstract_count}{$sent_count}{$phrase_count}{$map_count}{$term_count}{term}}{$mappings{$abstract_count}{$sent_count}{$phrase_count}{$map_count}{$term_count}{cui}}++;
	    
	  } # foreach
	  # last;
	} # foreach 
      } # foreach
      print "\n";
    } # foreach
    
    print "\n\n";

  } # foreach 
  

} # while




# Dump the dictionary file
open(DICT, ">$dict_file") or die "Can't open $dict_file for writing\n";
foreach $ambig_term (sort {$a cmp $b} keys %gold_standard_cuis) {
  print DICT "$ambig_term";
  foreach $cui (keys %{$gold_standard_cuis{$ambig_term}}) {
    print DICT " $gold_standard_cuis{$ambig_term}{$cui}";
  } # foreach 
  print DICT "\n";
} # foreach 

foreach $term (sort {$a cmp $b} keys %dictionary) {
  if(exists $gold_standard_cuis{$term}) {next; }
  print DICT "$term";
  foreach $cui (keys %{$dictionary{$term}}) {
    print DICT " $cui";
  } # foreach 
  print DICT "\n";
} # foreach 
close(DICT);


#####################################################################################################
#   SUBROUTINES
#####################################################################################################

sub norm_term { 
  
  my ($term) = @_;

  $term =~ s/\<[0-9]+\>//;
  $term =~ s/\, [A-z]+//;
  $term =~ tr/A-Z/a-z/;
  $term =~ s/(\/|\%|\{|\}|\(|\)|\[|\])//g;
  $term =~ s/^\*//;
  $term =~ s/^ +//;
  $term =~ s/ +$//;
  $term =~ s/ +/\_/g;

  return $term;

} # sub norm_term


sub usage {

  die("$0 -d dict_file -i dir -h\n\t-d(ictionary file):\t[defaults to \"Dictionary.txt\"]\n\t-i(nput directory):\t[defaults to \"OUT/\"]\n\t-s(top file):\t\tfile containing list of stop phrases\n\t-h(elp):\t\tprints out these usage notes\n\n");

} # sub usage
